# Copyright 2024 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: import <name> as <name> is required for names to be exported.
# See PEP 484 & https://github.com/jax-ml/jax/issues/7570

from jax._src.ffi import (
    build_ffi_lowering_function as build_ffi_lowering_function,
    ffi_call as ffi_call,
    ffi_lowering as ffi_lowering,
    include_dir as include_dir,
    pycapsule as pycapsule,
    register_ffi_target as register_ffi_target,
    register_ffi_type_id as register_ffi_type_id,
    register_ffi_type as register_ffi_type,
    register_ffi_target_as_batch_partitionable as register_ffi_target_as_batch_partitionable,
)
